from formencode.validators import Email
from pyramid.httpexceptions import HTTPFound
from pyramid.security import remember
from pyramid.security import forget

from kotti.security import get_principals
from kotti.resources import get_root
from kotti.resources import Node

def login(context, request):
    from kotti.views.util import TemplateAPIEdit
    root = get_root(request)
    api = TemplateAPIEdit(root, request)
    came_from = request.params.get('came_from', request.url)
    login, password = u'', u''
    if 'submitted' in request.params:
        login = request.params['login']
        password = request.params['password']
        principals = get_principals()
        principal = principals.get(login)
        if principal is None:
            try:
                Email().to_python(login)
            except Exception:
                pass
            else:
                results = [r for r in principals.search(email=login)]
                if results:
                    principal = results[0]

        if (principal is not None and
            principal.active and principal.confirm_token is None and
            principal.password == principals.hash_password(password)):
            headers = remember(request, login)
            request.session.flash(
                u"Welcome, %s!" % principal.title or principal.name, 'success')
            return HTTPFound(location=came_from, headers=headers)

        request.session.flash(u"Login failed.", 'error')

    return {
        'api': api,
        'url': request.application_url + '/@@login',
        'came_from': came_from,
        'login': login,
        'password': password,
        }

def logout(context, request):
    headers = forget(request)
    request.session.flash(u"You have been logged out.")
    location = request.params.get('came_from', request.application_url)
    return HTTPFound(location=location, headers=headers)

def includeme(config):
    config.add_view(
        login,
        context='pyramid.exceptions.Forbidden',
        renderer='../templates/edit/login.pt',
        )

    config.add_view(
        login,
        name='login',
        context=Node,
        renderer='../templates/edit/login.pt',
        )

    config.add_view(
        logout,
        name='logout',
        )



from pkg_resources import resource_filename
from pyramid.exceptions import Forbidden
from pyramid.httpexceptions import HTTPFound
from pyramid.security import has_permission
from pyramid.security import view_execution_permitted
from pyramid.url import resource_url
from pyramid.view import is_response
import colander
from deform import Form
from deform.widget import RichTextWidget
from deform.widget import TextAreaWidget

from kotti import configuration
from kotti.resources import DBSession
from kotti.resources import Node
from kotti.resources import Document
from kotti.security import get_principals
from kotti.security import map_principals_with_local_roles
from kotti.security import set_groups
from kotti.security import list_groups_raw
from kotti.security import list_groups_ext
from kotti.security import ROLES
from kotti.security import SHARING_ROLES
from kotti.views.util import TemplateAPIEdit
from kotti.views.util import addable_types
from kotti.views.util import title_to_name
from kotti.views.util import disambiguate_name
from kotti.views.util import FormController

deform_templates = resource_filename('deform', 'templates')
kotti_templates = resource_filename('kotti', 'templates/edit/widgets')
search_path = (kotti_templates, deform_templates)
Form.set_zpt_renderer(search_path)

class NodeSchema(colander.MappingSchema):
    title = colander.SchemaNode(colander.String())
    description = colander.SchemaNode(
        colander.String(),
        widget=TextAreaWidget(cols=40, rows=5),
        missing=u"",
        )

class DocumentSchema(NodeSchema):
    body = colander.SchemaNode(
        colander.String(),
        widget=RichTextWidget(theme='advanced'),
        missing=u"",
        )

def add_node(context, request):
    """This view's responsibility is to present the user with a form
    where they can choose between locations to add to, and types of
    nodes to add, and redirect to the actual add form based on this
    information.
    """
    all_types = configuration['kotti.available_types']
    
    if request.POST:
        what, where = request.POST['what'], request.POST['where']
        session = DBSession()
        what = [t for t in all_types if t.type_info.name == what][0]
        where = session.query(Node).get(int(where))
        location = resource_url(where, request) + '@@' + what.type_info.add_view
        return HTTPFound(location=location)

    possible_parents, possible_types = addable_types(context, request)
    if len(possible_parents) == 1 and len(possible_parents[0]['factories']) == 1:
        # Redirect to the add form straight away if there's only one
        # choice of parents and addable types:
        parent = possible_parents[0]
        add_view = parent['factories'][0].type_info.add_view
        location = resource_url(parent['node'], request) + '@@' + add_view
        return HTTPFound(location=location)

    # Swap first and second possible parents if there's no content in
    # 'possible_parents[0]' yet.  This makes the parent then the
    # default choice in the form:
    api = TemplateAPIEdit(context, request)
    if not api.list_children() and len(possible_parents) > 1:
        possible_parents[0], possible_parents[1] = (
            possible_parents[1], possible_parents[0])

    return {
        'api': api,
        'possible_parents': possible_parents,
        'possible_types': possible_types,
        }

def move_node(context, request):
    """This view allows copying, cutting, pasting, deleting of
    'context' and reordering of children of 'context'.
    """
    P = request.POST
    session = DBSession()
    
    if 'copy' in P:
        request.session['kotti.paste'] = (context.id, 'copy')
        request.session.flash(u'%s copied.' % context.title, 'success')
        return HTTPFound(location=request.url)

    if 'cut' in P:
        request.session['kotti.paste'] = (context.id, 'cut')
        request.session.flash(u'%s cut.' % context.title, 'success')
        return HTTPFound(location=request.url)

    if 'paste' in P:
        id, action = request.session['kotti.paste']
        item = session.query(Node).get(id)
        if action == 'cut':
            if not has_permission('edit', item, request):
                raise Forbidden() # XXX testme
            item.__parent__.children.remove(item)
            context.children.append(item)
            del request.session['kotti.paste']
        elif action == 'copy':
            copy = item.copy()
            name = copy.name
            if not name: # for root
                name = title_to_name(copy.title)
            while name in context.keys():
                name = disambiguate_name(name)
            copy.name = name
            context.children.append(copy)
        request.session.flash(u'%s pasted.' % item.title, 'success')
        return HTTPFound(location=request.url)

    if 'order-up' in P or 'order-down' in P:
        up, down = P.get('order-up'), P.get('order-down')
        id = int(down or up)
        if up is not None: # pragma: no cover
            mod = -1
        else:
            mod = +1

        child = session.query(Node).get(id)
        index = context.children.index(child)
        context.children.pop(index)
        context.children.insert(index+mod, child)
        request.session.flash(u'%s reordered.' % child.title, 'success')

    if 'delete' in P and 'delete-confirm' in P:
        parent = context.__parent__
        redirect_elements = []
        if view_execution_permitted(parent, request, 'edit'):
            redirect_elements.append('edit')
        location = resource_url(parent, request, *redirect_elements)
        request.session.flash(u'%s deleted.' % context.title, 'success')
        parent.children.remove(context)
        return HTTPFound(location=location)

    if 'rename' in P:
        name = P['name']
        title = P['title']
        context.name = name
        context.title = title
        request.session.flash(u'Item renamed', 'success')
        location = resource_url(context, request, 'move')        
        return HTTPFound(location=location)

    return {
        'api': TemplateAPIEdit(context, request),
        }

def share_node(context, request):
    flash = request.session.flash
    principals = get_principals()
    available_roles = [ROLES[role_name] for role_name in SHARING_ROLES]

    if 'apply' in request.params:
        changed = False
        p_to_r = {}
        for name in request.params:
            if name.startswith('orig-role::'):
                # orig-role::* is hidden checkboxes that allow us to
                # see what checkboxes were in the form originally
                token, principal_name, role_name = name.split('::')
                if role_name not in SHARING_ROLES:
                    raise Forbidden()
                new_value = bool(request.params.get(
                    'role::%s::%s' % (principal_name, role_name)))
                if principal_name not in p_to_r:
                    p_to_r[principal_name] = set()
                if new_value:
                    p_to_r[principal_name].add(role_name)

        for principal_name, new_role_names in p_to_r.items():
            # We have to be careful with roles that aren't mutable here:
            orig_role_names = set(list_groups_raw(principal_name, context))
            orig_sharing_role_names = set(
                r for r in orig_role_names if r in SHARING_ROLES)
            if new_role_names != orig_sharing_role_names:
                changed = True
                final_role_names = orig_role_names - set(SHARING_ROLES)
                final_role_names |= new_role_names
                set_groups(principal_name, context, final_role_names)

        if changed:
            flash(u'Your changes have been applied.', 'success')
        else:
            flash(u'No changes made.', 'info')
        return HTTPFound(location=request.url)

    existing = map_principals_with_local_roles(context)
    def with_roles(entry):
        all_groups = entry[1][0]
        return [g for g in all_groups if g.startswith('role:')]
    existing = filter(with_roles, existing)
    seen = set([entry[0].name for entry in existing])

    entries = []

    if 'search' in request.params:
        query = '*%s*' % request.params['query']
        found = False
        for p in principals.search(name=query, title=query, email=query):
            found = True
            if p.name not in seen:
                entries.append((p, list_groups_ext(p.name, context)))
        if not found:
            flash(u'No users or groups found.', 'info')

    entries = existing + entries

    return {
        'api': TemplateAPIEdit(context, request),
        'entries': entries,
        'available_roles': available_roles,
        'principals_to_roles': map_principals_with_local_roles(context),
        }

def edit_document(context, request):
    form = Form(DocumentSchema(), buttons=('save', 'cancel'))
    rendered = FormController(form)(context, request)
    if is_response(rendered):
        return rendered
    return {
        'api': TemplateAPIEdit(context, request),
        'form': rendered,
        }

def add_document(context, request):
    api = TemplateAPIEdit(
        context, request,
        first_heading=u'<h1>Add document to <em>%s</em></h1>' % context.title)
    form = Form(DocumentSchema(), buttons=('save', 'cancel'))
    rendered = FormController(form, add=Document)(context, request)
    if is_response(rendered):
        return rendered
    return {
        'api': api,
        'form': rendered,
        }

def includeme(config):
    config.add_view(
        edit_document,
        context=Document,
        name='edit',
        permission='edit',
        renderer='../templates/edit/node.pt',
        )

    config.add_view(
        add_document,
        name=Document.type_info.add_view,
        permission='add',
        renderer='../templates/edit/node.pt',
        )



from sqlalchemy import engine_from_config
from pyramid.authentication import AuthTktAuthenticationPolicy
from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.config import Configurator
from pyramid.session import UnencryptedCookieSessionFactoryConfig
from pyramid.util import DottedNameResolver

class Configuration(dict):
    """A dict that can resolve dotted names to Python objects the
    first time they're accessed.
    """
    def __init__(self, d, dotted_names):
        super(Configuration, self).__init__(d)
        self.dotted_names = dotted_names

    def __getitem__(self, key):
        value = super(Configuration, self).__getitem__(key)
        if key in self.dotted_names and isinstance(value, basestring):
            values = []
            for dottedname in value.split():
                values.append(DottedNameResolver(None).resolve(dottedname))
            super(Configuration, self).__setitem__(key, values)
            return values
        else:
            return value

    def __getattr__(self, key): # pragma: no cover
        return self[key]

def authtkt_factory(**kwargs):
    from kotti.security import list_groups_callback
    kwargs.setdefault('callback', list_groups_callback)
    return AuthTktAuthenticationPolicy(**kwargs)

def acl_factory(**kwargs):
    return ACLAuthorizationPolicy()

def cookie_session_factory(**kwargs):
    return UnencryptedCookieSessionFactoryConfig(**kwargs)

def none_factory(**kwargs): # pragma: no cover
    return None

# All of these can be set by passing them in the Paste Deploy settings:
configuration = Configuration(
    {
        'kotti.templates.master_view': 'kotti:templates/view/master.pt',
        'kotti.templates.master_edit': 'kotti:templates/edit/master.pt',
        'kotti.templates.master_cp': 'kotti:templates/site-setup/master.pt',
        'kotti.templates.base_css': 'kotti:static/base.css',
        'kotti.templates.view_css': 'kotti:static/view.css',
        'kotti.templates.edit_css': 'kotti:static/edit.css',
        'kotti.configurators': '',
        'kotti.includes': 'kotti.events kotti.views.view kotti.views.edit kotti.views.login kotti.views.site_setup',
        'kotti.available_types': 'kotti.resources.Document',
        'kotti.authn_policy_factory': 'kotti.authtkt_factory',
        'kotti.authz_policy_factory': 'kotti.acl_factory',
        'kotti.session_factory': 'kotti.cookie_session_factory',
        'kotti.principals': 'kotti.security.principals',
    },
    dotted_names=set([
        'kotti.configurators',
        'kotti.includes',
        'kotti.available_types',
        'kotti.authn_policy_factory',
        'kotti.authz_policy_factory',
        'kotti.session_factory',
        'kotti.principals',
        ]),
    )

def main(global_config, **settings):
    """ This function returns a WSGI application.
    """
    for key in configuration:
        if key in settings:
            configuration[key] = settings.pop(key)

    # Allow extending packages to change 'configuration' w/ Python:
    for func in configuration['kotti.configurators']:
        func(configuration)

    secret1 = settings['kotti.secret']
    secret2 = settings.get('kotti.secret2', secret1)
    configuration.secret = secret1
    configuration.secret2 = secret2    

    from kotti.resources import appmaker
    engine = engine_from_config(settings, 'sqlalchemy.')
    get_root = appmaker(engine)

    authentication_policy = configuration[
        'kotti.authn_policy_factory'][0](secret=secret1)
    authorization_policy = configuration[
        'kotti.authz_policy_factory'][0]()
    session_factory = configuration['kotti.session_factory'][0](secret=secret2)

    config = Configurator(
        settings=settings,
        root_factory=get_root,
        authentication_policy=authentication_policy,
        authorization_policy=authorization_policy,
        session_factory=session_factory,
        )

    _configure_base_views(config)

    # Include modules listed in 'includeme' configuration:
    for module in configuration['kotti.includes']:
        config.include(module)

    return config.make_wsgi_app()

def _configure_base_views(config):
    from kotti.resources import Node

    config.add_static_view('static-deform', 'deform:static')
    config.add_static_view('static-kotti', 'kotti:static')
    config.add_view('kotti.views.view.view_node_default', context=Node)
    config.add_view(
        'kotti.views.edit.add_node',
        name='add',
        permission='add',
        renderer='templates/edit/add.pt',
        )
    config.add_view(
        'kotti.views.edit.move_node',
        name='move',
        permission='edit',
        renderer='templates/edit/move.pt',
        )
    config.add_view(
        'kotti.views.edit.share_node',
        name='share',
        permission='manage',
        renderer='templates/edit/share.pt',
        )



import string

from pyramid.decorator import reify
from pyramid.httpexceptions import HTTPFound
from pyramid.location import inside
from pyramid.location import lineage
from pyramid.renderers import get_renderer
from pyramid.security import authenticated_userid
from pyramid.security import has_permission
from pyramid.security import view_execution_permitted
from pyramid.url import resource_url
from deform import ValidationFailure

from kotti import configuration
from kotti.resources import DBSession
from kotti.resources import Node
from kotti.security import get_principals

class TemplateAPI(object):
    """This implements the 'api' object that's passed to all
    templates.

    Use dict-access as a shortcut to retrieve template macros from
    templates.  ``api['master_edit.messages']`` will return the
    'messages' macro from the 'master_edit' template.
    """
    macro_templates = dict(
        master_view=configuration['kotti.templates.master_view'],
        master_edit=configuration['kotti.templates.master_edit'],
        master_cp=configuration['kotti.templates.master_cp'],
        )

    base_css = configuration['kotti.templates.base_css']
    view_css = configuration['kotti.templates.view_css']
    edit_css = configuration['kotti.templates.edit_css']

    def __init__(self, context, request, **kwargs):
        self.context, self.request = context, request
        self.__dict__.update(kwargs)

    def __getitem__(self, dottedname):
        try:
            template_name, macro_name = dottedname.split('.')
        except ValueError: # Chameleon will try dict access after attr access
            raise KeyError(dottedname)
        
        template = self.macro_templates[template_name]
        if isinstance(template, basestring):
            template = self.macro_templates[template_name] = get_renderer(
                template).implementation()
        return template.macros[macro_name]

    @reify
    def page_title(self):
        return u'%s - %s' % (self.context.title, self.root.title)

    def url(self, context=None, *elements):
        if context is None:
            context = self.context
        rhs = '/'.join(elements)
        return resource_url(context, self.request) + rhs

    @reify
    def root(self):
        return self.lineage[-1]

    @reify
    def lineage(self):
        return list(lineage(self.context))

    @reify
    def user(self):
        userid = authenticated_userid(self.request)
        return get_principals().get(userid)
    
    def has_permission(self, permission, context=None):
        if context is None:
            context = self.context
        return has_permission(permission, context, self.request)

    def list_children(self, context=None, permission='view'):
        if context is None:
            context = self.context
        children = []
        for child in context.values():
            if (not permission or
                has_permission(permission, child, self.request)):
                children.append(child)
        return children

    def list_children_go_up(self, context=None, permission='view'):
        if context is None:
            context = self.context
        children = self.list_children(context, permission)
        if not children and context.__parent__ is not None:
            children = self.list_children(context.__parent__)
        return children

    inside = staticmethod(inside)

class TemplateAPIEdit(TemplateAPI):
    @reify
    def page_title(self):
        return u'%s - %s' % (
            self.request.view_name.replace('_', ' ').title(), self.root.title)

    @reify
    def first_heading(self):
        return u'<h1>Edit <em>%s</em></h1>' % self.context.title

    def _find_edit_view(self, item):
        view_name = self.request.view_name
        if not view_execution_permitted(item, self.request, view_name):
            view_name = u'edit' # XXX testme
        if not view_execution_permitted(item, self.request, view_name):
            view_name = u'' # XXX testme
        return view_name

    def _make_links(self, items):
        links = []
        for item in items:
            view_name = self._find_edit_view(item)
            url = resource_url(item, self.request, view_name)
            links.append(dict(
                url=url,
                name=item.title,
                is_edit_link=view_name != '',
                node=item,
                is_context=item == self.context,
                ))
        return links

    @reify
    def breadcrumbs(self):
        return self._make_links(tuple(reversed(self.lineage)))

    @reify
    def context_links(self):
        siblings = []
        if self.context.__parent__ is not None:
            siblings = self._make_links(
                self.list_children(self.context.__parent__))
            siblings = filter(lambda l: not l['is_context'], siblings)
        children = self._make_links(self.list_children(self.context))
        return siblings, children

    @reify
    def edit_links(self):
        return [l for l in self.context.type_info.edit_links
                if l.permitted(self.context, self.request)]

    def more_links(self, name):
        return [l for l in getattr(self, name)
                if l.permitted(self.context, self.request)]

    def get_paste_item(self):
        info = self.request.session.get('kotti.paste')
        if info:
            id, action = info
            item = DBSession().query(Node).get(id)
            if not item.type_info.addable(self.context, self.request):
                return
            if action == 'cut' and self.inside(self.context, item):
                return
            if self.context == item:
                return
            return item

def addable_types(context, request):
    all_types = configuration['kotti.available_types']

    # 'possible_parents' is a list of dicts with 'node' and 'factories',
    # where 'node' is the context to add to and 'factories' is the list
    # of factories that can be applied in the contetx:
    possible_parents = []
    parent = context
    while parent is not None:
        possible_parents.append({'node': parent, 'factories': []})
        parent = parent.__parent__

    for entry in possible_parents:
        parent = entry['node']
        for factory in all_types:
            if factory.type_info.addable(parent, request):
                entry['factories'].append(factory)

    possible_parents = filter(lambda e: e['factories'], possible_parents)

    _possible_types = {}
    for entry in possible_parents:
        for factory in entry['factories']:
            name = factory.type_info.name
            pt = _possible_types.get(
                name, {'factory': factory, 'nodes': []})
            pt['nodes'].append(entry['node'])
            _possible_types[name] = pt

    possible_types = []
    for t in all_types:
        entry = _possible_types.get(t.type_info.name)
        if entry:
            possible_types.append(entry)

    return possible_parents, possible_types

def title_to_name(title):
    okay = string.letters + string.digits + '-'
    name = u'-'.join(title.lower().split())
    name = u''.join(ch for ch in name if ch in okay)
    return name

def disambiguate_name(name):
    parts = name.split(u'-')
    if len(parts) > 1:
        try:
            index = int(parts[-1])
        except ValueError:
            parts.append(u'1')
        else:
            parts[-1] = unicode(index+1)
    else:
        parts.append(u'1')
    return u'-'.join(parts)

class FormController(object):
    add = None
    post_key = 'save'
    edit_success_msg = u"Your changes have been saved."
    add_success_msg = u"Successfully added item."
    error_msg = (u"There was a problem with your submission.\n"
                 u"Errors have been highlighted below.")
    success_path = 'edit'

    def __init__(self, form, **kwargs):
        self.form = form
        for key, value in kwargs.items():
            if key in self.__class__.__dict__:
                setattr(self, key, value)
            else: # pragma: no coverage
                raise TypeError("Unknown argument %r" % key)

    def __call__(self, context, request):
        if self.post_key in request.POST:
            controls = request.POST.items()
            try:
                appstruct = self.form.validate(controls)
            except ValidationFailure, e:
                request.session.flash(self.error_msg, 'error')
                return e.render()
            else:
                if self.add is None: # edit
                    for key, value in appstruct.items():
                        setattr(context, key, value)
                    request.session.flash(self.edit_success_msg, 'success')
                    location = resource_url(context, request, self.success_path)
                    return HTTPFound(location=location)
                else: # add
                    name = title_to_name(appstruct['title'])
                    while name in context.keys():
                        name = disambiguate_name(name)
                    item = context[name] = self.add(**appstruct)
                    request.session.flash(self.add_success_msg, 'success')
                    location = resource_url(item, request, self.success_path)
                    return HTTPFound(location=location)
        else: # no post means less action
            if self.add is None:
                return self.form.render(context.__dict__)
            else:
                return self.form.render()

def is_root(context, request):
    return context is TemplateAPI(context, request).root
